/*****************************************************************************

       Copyright   1995, 1996 Digital Equipment Corporation,
                       Maynard, Massachusetts.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, provided  
that the copyright notice and this permission notice appear in all copies  
of software and supporting documentation, and that the name of Digital not  
be used in advertising or publicity pertaining to distribution of the software 
without specific, written prior permission. Digital grants this permission 
provided that you prominently mark, as not part of the original, any 
modifications made to this software or documentation.

Digital Equipment Corporation disclaims all warranties and/or guarantees  
with regard to this software, including all implied warranties of fitness for 
a particular purpose and merchantability, and makes no representations 
regarding the use of, or the results of the use of, the software and 
documentation in terms of correctness, accuracy, reliability, currentness or
otherwise; and you rely on the software, documentation and results solely at 
your own risk. 

******************************************************************************/

/*****************************************************************************
*
* Copyright 1996, 1997, 1998 Hans Reiser, see /usr/src/linux/reiserfs/README
*                     for licensing and copyright details
*
*****************************************************************************/


/* This is a set of functions that provides minimal filesystem
 * functionality to the Linux bootstrapper.  All we can do is
 * open and read files... but that's all we need 8-)
 */

//#define __KERNEL__

//#define _ALPHA_CURRENT_H

#include <linux/kernel.h>
#include <linux/config.h>
#include <linux/string.h>
#include <linux/stat.h>
#include <linux/kdev_t.h>



/*>>>>>>>>>>>>*/
#include <linux/sched.h>
#include <asm/uaccess.h>
#include <linux/locks.h>
#include <linux/init.h>
#include <unistd.h>
#include <linux/unistd.h>
#include <string.h>
/*<<<<<<<<<<<<*/

#define DEBUG_REISERFS


#include <linux/reiserfs_fs.h>
#include <linux/reiserfs_fs_sb.h>

#include "fs.h"

#define REISERFS_BLOCK_SIZE 4096




/*
 * Structure of an inode on the disk
 */
struct reiserfs_inode
{

	__u32 i_key [4];
	__u32 i_first_direct_byte;
	int i_data_length;
	__u32 i_data [2];
//	__u32 i_data [REISERFS_N_BLOCKS];
	
//	MAX_ITEM_LEN(REISERFS_BLOCK_SIZE)/sizeof(__u32)];

	unsigned short i_mode;	/* File mode */
	unsigned short i_uid;	/* Owner Uid */
	unsigned int i_size;	/* Size in bytes */
	unsigned int i_atime;	/* Access time */
	unsigned int i_ctime;	/* Creation time */
	unsigned int i_mtime;	/* Modification time */
	unsigned int i_dtime;	/* Deletion Time */
	unsigned short i_gid;	/* Group Id */
	unsigned short i_links_count;	/* Links count */
	unsigned int i_blocks;	/* Blocks count */
	unsigned int i_flags;	/* File flags */
	unsigned int i_version;	/* File version (for NFS) */
};

struct milo_item
{
	struct key ih_key;
	union
	{
		__u16 ih_free_space; 
		__u16 ih_entry_count; 
	} u;
	__u16 ih_item_len;
	__u16 ih_item_location;
	__u16 ih_reserved;
	__u16 it_Last;
	__u16 it_Number;
	struct key ih_r_key;
	union
	{
		struct stat_data sd;
		struct reiserfs_de_head deh[(MAX_ITEM_LEN(REISERFS_BLOCK_SIZE)-DEH_SIZE)/DEH_SIZE];
		__u32 item[MAX_ITEM_LEN(REISERFS_BLOCK_SIZE)/sizeof(__u32)];
		char txt[MAX_ITEM_LEN(REISERFS_BLOCK_SIZE)];
	} body;
} ;


struct reiserfs_journal_ptr
{
	__u32 r_BlockNr;
	__u32 j_BlockNr;
};

struct reiserfs_journal_ptr j_table[REISERFS_BLOCK_SIZE];

__u32 j_tab_count;

struct reiserfs_journal_desc desc;
struct reiserfs_journal_commit commit;
struct reiserfs_journal_header jh ;


#define MAX_OPEN_FILES		4

int fd ;

unsigned long ErrorNumber;

static struct milo_item n_item_buf;
static struct milo_item it_buf;
static struct milo_item it_open;

static struct item_head * ih;
static struct block_head * bh;
static struct reiserfs_super_block sb;

static struct key root_key = {REISERFS_ROOT_PARENT_OBJECTID, REISERFS_ROOT_OBJECTID, 0, 0};
static hashf_t p_hash_function;


#define comp_keys(k1,k2)        __comp_keys((k1), (k2), REISERFS_FULL_KEY_LEN)
#define comp_short_keys(k1,k2)  __comp_keys((k1), (k2), REISERFS_SHORT_KEY_LEN)

#define VV_ITEMS(bh)	  	   ((__u32)((bh)->blk_nr_item))
#define VV_CHILD(bh,n_pos)     ((struct disk_child *)((bh)+BLKH_SIZE+VV_ITEMS(bh)*KEY_SIZE+DC_SIZE*(n_pos)))
#define VV_CHILD_NUM(bh,n_pos) ((__u32)VV_CHILD((bh),(n_pos))->dc_block_number)
#define VV_KEY(bh,n_pos)       ((struct key *)((char *)(bh)+BLKH_SIZE+KEY_SIZE*(n_pos)))
#define VV_IHEAD(bh,n_pos)     ((struct item_head *)((char *)(bh)+BLKH_SIZE+IH_SIZE*(n_pos)))
#define VV_ITEM(bh,n_pos)      ((void *)((char *)(bh)+(VV_IHEAD((bh),(n_pos)))->ih_item_location))
#define VV_ITEM_SIZE(bh,n_pos) ((unsigned short)((bh)+(VV_IHEAD((bh),(n_pos)))->ih_item_len))

#define VV_M_IHEAD(p_i,bh,n_pos) vv_m_ihead((p_i),(bh),(n_pos))



static int blocksize;

//atic char filename[256];

static char blkbuf[REISERFS_BLOCK_SIZE];
static char iblkbuf[REISERFS_BLOCK_SIZE];

static struct inode_table_entry
{
	struct reiserfs_inode inode;
	int inumber;
	int free;
	unsigned short old_mode;
} inode_table[MAX_OPEN_FILES];

static struct reiserfs_inode *reiserfs_m_iget( struct key * ino_k, struct milo_item * tmp_item);
static int reiserfs_m_bread(int fd, long blkno, char *buffer);
static void reiserfs_m_breadi(struct reiserfs_inode *ip, long blkno, char *buffer);
static int reiserfs_m_mount(long device, int quiet);
static int reiserfs_m_open(char *filename);
static void reiserfs_m_cat(struct reiserfs_inode *ip, char *buffer);
static void reiserfs_m_ls(char *dirname, char *devname);
static void reiserfs_m_close(int fd);
static long reiserfs_m_fdsize(int fd);



#define bread(a,b,c) mm_bread((a),(b),(c))
#if 0
static inline struct buffer_head * mm_bread(int dev, __u32 blkno, __u32 b_size)
{
	char * blk_bf;
	struct buffer_head *bh;
	bh=kmalloc(sizeof( struct buffer_head) + blocksize ,GFP_KERNEL);
	blk_bf=(char*)bh + sizeof(struct buffer_head);
	bh->b_data=blk_bf;
	bh->b_blocknr=blkno;
	if (device_read(dev, blk_bf, b_size, blkno*b_size) !=  blkno*b_size)
		{
			printk("\nBread(inline) error on read blkno%d",blkno);
			return NULL;
		}
	return bh;
}
#endif
static inline struct buffer_head * mm_bread(int dev, __u32 blkno, __u32 b_size)
{
	char * blk_bf;
	struct buffer_head *bh;
	bh=kmalloc(sizeof( struct buffer_head) + blocksize ,GFP_KERNEL);
	blk_bf=(char*)bh + sizeof(struct buffer_head);
	bh->b_data=blk_bf;
	bh->b_blocknr=blkno;
	if (device_read(dev, blk_bf, b_size, blkno*b_size) !=  blkno*b_size)
		{
			printk("\nBread(inline) error on read blkno%d",blkno);
			return NULL;
		}
	return bh;
}







#define brelse(a) mm_brelse((a))

static inline __u32 mm_brelse(struct buffer_head *bh)
{
	kfree(bh);
}

struct bootfs reiserfs =
{
	"reiserfs",
	0,
	0,

	reiserfs_m_mount,
	reiserfs_m_open,
	reiserfs_m_fdsize,
	reiserfs_m_bread,
	reiserfs_m_ls,
	reiserfs_m_close,
};

/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/




/* compares description block with commit block.  returns 1 if they differ, 0 if they are the same */
static int journal_compare_desc_commit(struct reiserfs_super_block *p_s_sb,
									   struct reiserfs_journal_desc *desc, 
									   struct reiserfs_journal_commit *commit)
{
	if (commit->j_trans_id != desc->j_trans_id ||
		commit->j_len != desc->j_len ||
		commit->j_len > JOURNAL_TRANS_MAX || 
		commit->j_len <= 0 )
		{
			return 1 ;
		}
  return 0 ;
}

/* returns 0 if it did not find a description block  
** returns -1 if it found a corrupt commit block
** returns 1 if both desc and commit were valid 
*/
static int journal_transaction_is_valid(struct reiserfs_super_block *p_s_sb,
										struct buffer_head *d_bh,
										__u32 *oldest_invalid_trans_id,
										__u32 *newest_mount_id)
{
	struct reiserfs_journal_desc *desc ;
	struct reiserfs_journal_commit *commit ;

	struct reiserfs_super_block *rs = &sb ;

	struct buffer_head *c_bh ;
	__u32 offset ;

	if (!d_bh) 
		return 0 ;

	desc = (struct reiserfs_journal_desc *)d_bh->b_data ;

	if (desc->j_len > 0 && !memcmp(desc->j_magic, JOURNAL_DESC_MAGIC, 8))
		{
			if (oldest_invalid_trans_id &&
				*oldest_invalid_trans_id && 
				desc->j_trans_id > *oldest_invalid_trans_id)
				{
					return 0 ;
				}
			if (newest_mount_id && *newest_mount_id > desc->j_mount_id)
				{
					return -1 ;
				}
			offset = d_bh->b_blocknr - rs->s_journal_block ;
		
			/* ok, we have a journal description block, lets see if the 
			** transaction was valid 
			*/
			commit = bread(s_dev,
						   rs->s_journal_block + ((offset + desc->j_len + 1) % JOURNAL_BLOCK_COUNT), 
						   rs->s_blocksize) ;
			if (!c_bh)
				return 0 ;
			
				commit = (struct reiserfs_journal_commit *)c_bh->b_data ;
			if (journal_compare_desc_commit(p_s_sb, desc, commit))
				{
					brelxse(commit) ;
					if (oldest_invalid_trans_id)
						*oldest_invalid_trans_id = desc->j_trans_id ;
					return -1; 
				}
			brelse(c_bh) ;
			return 1 ;
		}
	else
		{
			return 0 ;
		}
}




static int journal_read_transaction(struct reiserfs_super_block *p_s_sb,
									__u32 cur_dblock,
									__u32 oldest_start, 
									__u32 oldest_trans_id,
									__u32 newest_mount_id)
{
	struct reiserfs_journal_desc *desc ;
	struct reiserfs_journal_commit *commit ;
	struct reiserfs_super_block *rs = &sb ;
	__u32 trans_id = 0 ;
	struct buffer_head *c_bh ;
	struct buffer_head *d_bh ;
	struct buffer_head **log_blocks = NULL ;
	struct buffer_head **real_blocks = NULL ;
	__u32 trans_offset ;
	int i;

	d_bh = bread(s_dev, cur_dblock, p_s_sb->s_blocksize) ;
	if (!d_bh)
		return 1 ;

	desc = (struct reiserfs_journal_desc *)d_bh->b_data ;
	trans_offset = d_bh->b_blocknr - rs->s_journal_block ;

	if (desc->j_trans_id < oldest_trans_id)
		{
			brelse(d_bh) ;
			return 1 ;
		}
	if (desc->j_mount_id != newest_mount_id)
		{
			brelse(d_bh) ;
			return 1 ;
		}
	c_bh = bread(s_dev, rs->s_journal_block + ((trans_offset + desc->j_len + 1) % JOURNAL_BLOCK_COUNT), 
				 p_s_sb->s_blocksize) ;
	if (!c_bh)
		{
			brelse(d_bh) ;
			return 1 ;
		}
	commit = (struct reiserfs_journal_commit *)c_bh->b_data ;
	if (journal_compare_desc_commit(p_s_sb, desc, commit))
		{
			brelse(c_bh) ;
			brelse(d_bh) ;
			return 1; 
		}
	trans_id = desc->j_trans_id ;

                            /* now we know we've got a good transaction, and it was inside the valid time ranges */
	log_blocks = kmalloc(desc->j_len * sizeof(struct buffer_head *), GFP_KERNEL) ;
	//real_blocks = kmalloc(desc->j_len * sizeof(struct buffer_head *), GFP_KERNEL) ;
	if (!log_blocks )
		{
			brelse(c_bh) ;
			brelse(d_bh) ;
			kfree(log_blocks) ;
//			kfree(real_blocks) ;
			reiserfs_warning("clm-2025: kmalloc failed, unable to mount FS\n") ;
			return -1 ;
		}
	/* get all the buffer heads */
	for(i = 0 ; i < desc->j_len ; i++)
		{
			log_blocks[i] = getblk(s_dev,
								   rs->s_journal_block + (trans_offset + 1 + i) % JOURNAL_BLOCK_COUNT,
								   rs->s_blocksize);
			if (i < JOURNAL_TRANS_HALF)
				{
					real_blocks[i] = getblk(s_dev, desc->j_realblock[i], p_s_sb->s_blocksize) ;
				}
			else
				{
					real_blocks[i] = getblk(s_dev, commit->j_realblock[i - JOURNAL_TRANS_HALF],
											p_s_sb->s_blocksize) ;
				}
			if (real_blocks[i]->b_blocknr >= rs->s_journal_block &&
				real_blocks[i]->b_blocknr < (rs->s_journal_block+JOURNAL_BLOCK_COUNT))
				{
					reiserfs_warning("clm-2026: REPLAY FAILURE fsck required! Trying to replay onto a log block\n") ;
					brelse_array(log_blocks, i) ;
					brelse_array(real_blocks, i) ;
					brelse(c_bh) ;
					brelse(d_bh) ;
					kfree(log_blocks) ;
					kfree(real_blocks) ;
					return -1 ;
				}
		}
	/* read in the log blocks, memcpy to the corresponding real block */
	ll_rw_block(READ, desc->j_len, log_blocks) ;

	for (i = 0 ; i < desc->j_len ; i++)
		{
			memcpy(real_blocks[i]->b_data, log_blocks[i]->b_data, real_blocks[i]->b_size) ;
			mark_buffer_uptodate(real_blocks[i], 1) ;
			brelse(log_blocks[i]) ;
		}

	cur_dblock = rs->s_journal_block + ((trans_offset + desc->j_len + 2) % JOURNAL_BLOCK_COUNT) ;


	/* init starting values for the first transaction, in case this is the last transaction to be replayed. */
	j_start = cur_dblock - rs->s_journal_block ;
	j_last_flush_trans_id = trans_id ;
	j_trans_id = trans_id + 1;
	brelse(c_bh) ;
	brelse(d_bh) ;
	kfree(log_blocks) ;
	kfree(real_blocks) ;
	return 0 ;
}


/*
** read and replay the log
** on a clean unmount, the journal header's next unflushed pointer will be to an invalid
** transaction.  This tests that before finding all the transactions in the log, whic makes normal mount times fast.
**
** After a crash, this starts with the next unflushed transaction, and replays until it finds one too old, or invalid.
**
** On exit, it sets things up so the first transaction will work correctly.
*/
static int journal_read(struct super_block *p_s_sb)
{
	struct reiserfs_super_block *rs = &sb;
	struct reiserfs_journal_desc *desc ;
	__u32 last_flush_trans_id = 0 ;
	__u32 oldest_trans_id = 0;
	__u32 oldest_invalid_trans_id = 0 ;
	time_t start ;
	__u32 last_flush_start = 0;
	__u32 oldest_start = 0;
	__u32 cur_dblock = 0 ;
	__u32 newest_mount_id = 9 ;
	struct buffer_head *d_bh ;
	struct reiserfs_journal_header *jh ;
	int valid_journal_header = 0 ;
	int replay_count = 0 ;
	int continue_replay = 1 ;
	int ret ;

	cur_dblock = rs->s_journal_block ;
	start = CURRENT_TIME ;

	/* step 1, read in the journal header block.  Check the transaction is says is the first unflushed,
	** and if that transaction is not valid, replay is done
	*/
	jh;
	d_bh = bread(p_s_sb->s_dev, rs->s_journal_block +	JOURNAL_BLOCK_COUNT,
				 p_s_sb->s_blocksize) ;
	jh = d_bh->b_data.
	if (jh->j_first_unflushed_offset >= 0 &&
		jh->j_first_unflushed_offset < JOURNAL_BLOCK_COUNT &&
		jh->j_last_flush_trans_id > 0)
		{
			last_flush_start = rs->s_journal_block + jh->j_first_unflushed_offset ;
			last_flush_trans_id = jh->j_last_flush_trans_id ;
			valid_journal_header = 1 ;
			
			/* now, we try to read the first unflushed offset.  If it is not valid, there is nothing more we can do,
			** and it makes no sense to read through the whole log.
			*/


			d_bh = bread(p_s_sb->s_dev, last_flush_start, p_s_sb->s_blocksize) ;
			ret = journal_transaction_is_valid(p_s_sb, d_bh, NULL, NULL) ;

			if (!ret)
				{
					continue_replay = 0 ;
				}
			if (d_bh)
				{
					brelse(d_bh) ;
				}
		}
	
	/* ok, there are transactions that need to be replayed.  
	** start with the first log block, find all the valid transactions, 
	** and pick out the oldest.
	*/
	while(continue_replay  &&   cur_dblock < (rs->s_journal_block + JOURNAL_BLOCK_COUNT) )
		{
			d_bh = bread(p_s_sb->s_dev, cur_dblock, p_s_sb->s_blocksize) ;
		if ((ret = journal_transaction_is_valid(p_s_sb, d_bh, 
												&oldest_invalid_trans_id, 
												&newest_mount_id)) == 1)
			{
				desc = (struct reiserfs_journal_desc *)d_bh->b_data ;
				if ( oldest_start == 0 || newest_mount_id < desc->j_mount_id)
					{
						oldest_trans_id = desc->j_trans_id ;
						oldest_start = d_bh->b_blocknr ;
						newest_mount_id = desc->j_mount_id ;
					}
				else
					{
						if (oldest_trans_id > desc->j_trans_id)  /* one just read was older */
							{
								oldest_trans_id = desc->j_trans_id ;
								oldest_start = d_bh->b_blocknr ;
							}
					}
				cur_dblock += desc->j_len + 2 ;
			} 
		else
			{
				cur_dblock++ ;
			}
		if (d_bh)
			{
				brelse(d_bh) ;
			}
		}
	/* step three, starting at the oldest transaction, replay */
	if (last_flush_start > 0)
		{
			oldest_start = last_flush_start ;
			oldest_trans_id = last_flush_trans_id ;
		} 
	cur_dblock = oldest_start ;

	if (oldest_trans_id)
		{
			printk("clm-2035: Starting replay from offset %lu, trans_id %lu\n",
				   cur_dblock - rs->s_journal_block, oldest_trans_id) ;
		}

	replay_count = 0 ;
	while(continue_replay && oldest_trans_id > 0)
		{
			ret = journal_read_transaction(p_s_sb, cur_dblock, oldest_start, oldest_trans_id, newest_mount_id) ;
			if (ret < 0)
				{
					brelse(SB_JOURNAL(p_s_sb)->j_header_bh);
					SB_JOURNAL(p_s_sb)->j_header_bh = NULL ;
					return ret ;
				}
			else
				if (ret != 0)        /* ==1 */
					{
						break ;
					}
			cur_dblock = rs->s_journal_block + SB_JOURNAL(p_s_sb)->j_start ;
			replay_count++ ;
		}

	if (oldest_trans_id == 0)
		{
			printk("clm-2036: No valid transactions found\n") ;
		}

	printk("clm-2038: Replayed %d transactions in %lu seconds, mount_id now %lu\n",
		   replay_count, CURRENT_TIME - start, SB_JOURNAL(p_s_sb)->j_mount_id) ;

	return 0 ;
}



/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<---------------->>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/



inline int  __comp_keys (void * k1, void * k2, int key_size)
{
  __u32 * p_s_key1, * p_s_key2;
  int n_key_length = key_size;
  p_s_key1 = (__u32 *)k1;
  p_s_key2 = (__u32 *)k2;
  for( ; n_key_length--; ++p_s_key1, ++p_s_key2 ) 
	  {
	  if ( *p_s_key1 < *p_s_key2 )  return -1;
	  if ( *p_s_key1 > *p_s_key2 )  return 1;
	  }
  return 0;
}


#define  MAX_GEN_NUMBER 127


#include <asm/types.h>

#define DELTA 0x9E3779B9
#define FULLROUNDS 10		/* 32 is overkill, 16 is strong crypto */
#define PARTROUNDS 6		/* 6 gets complete mixing */

#ifndef __KERNEL__
typedef __u32 u32;
#endif

/* a, b, c, d - data; h0, h1 - accumulated hash */
#define TEACORE(rounds)							\
	do {								\
		u32 sum = 0;						\
		int n = rounds;						\
		u32 b0, b1;						\
									\
		b0 = h0;						\
		b1 = h1;						\
									\
		do							\
		{							\
			sum += DELTA;					\
			b0 += ((b1 << 4)+a) ^ (b1+sum) ^ ((b1 >> 5)+b);	\
			b1 += ((b0 << 4)+c) ^ (b0+sum) ^ ((b0 >> 5)+d);	\
		} while(--n);						\
									\
		h0 += b0;						\
		h1 += b1;						\
	} while(0)


static u32 keyed_hash(const char *msg, int len)
{
	u32 k[] = { 0x9464a485, 0x542e1a94, 0x3e846bff, 0xb75bcfc3}; 

	u32 h0 = k[0], h1 = k[1];
	u32 a, b, c, d;
	u32 pad;
	int i;
 

	//	assert(len >= 0 && len < 256);

	pad = (u32)len | ((u32)len << 8);
	pad |= pad << 16;

	while(len >= 16)
		{
			a = (u32)msg[ 0]      |
				(u32)msg[ 1] << 8 |
				(u32)msg[ 2] << 16|
				(u32)msg[ 3] << 24;
			b = (u32)msg[ 4]      |
				(u32)msg[ 5] << 8 |
				(u32)msg[ 6] << 16|
				(u32)msg[ 7] << 24;
			c = (u32)msg[ 8]      |
				(u32)msg[ 9] << 8 |
				(u32)msg[10] << 16|
				(u32)msg[11] << 24;
			d = (u32)msg[12]      |
				(u32)msg[13] << 8 |
				(u32)msg[14] << 16|
				(u32)msg[15] << 24;
		
			TEACORE(PARTROUNDS);

			len -= 16;
		msg += 16;
		}
	
	if (len >= 12)
		{
	    	//assert(len < 16);
			if (len >= 16)
				*(int *)0 = 0;

			a = (u32)msg[ 0]      |
				(u32)msg[ 1] << 8 |
				(u32)msg[ 2] << 16|
				(u32)msg[ 3] << 24;
			b = (u32)msg[ 4]      |
				(u32)msg[ 5] << 8 |
				(u32)msg[ 6] << 16|
				(u32)msg[ 7] << 24;
			c = (u32)msg[ 8]      |
				(u32)msg[ 9] << 8 |
				(u32)msg[10] << 16|
				(u32)msg[11] << 24;
			
			d = pad;
			for(i = 12; i < len; i++)
				{
					d <<= 8;
					d |= msg[i];
				}
		}
	else
		if (len >= 8)
			{
				//assert(len < 12);
				if (len >= 12)
					*(int *)0 = 0;
				a = (u32)msg[ 0]      |
					(u32)msg[ 1] << 8 |
					(u32)msg[ 2] << 16|
					(u32)msg[ 3] << 24;
				b = (u32)msg[ 4]      |
					(u32)msg[ 5] << 8 |
					(u32)msg[ 6] << 16|
					(u32)msg[ 7] << 24;

				c = d = pad;
				for(i = 8; i < len; i++)
					{
						c <<= 8;
						c |= msg[i];
					}
			}
		else
			if (len >= 4)
				{
					//assert(len < 8);
					if (len >= 8)
						*(int *)0 = 0;
					a = (u32)msg[ 0]      |
						(u32)msg[ 1] << 8 |
						(u32)msg[ 2] << 16|
						(u32)msg[ 3] << 24;
					
					b = c = d = pad;
					for(i = 4; i < len; i++)
						{
							b <<= 8;
							b |= msg[i];
						}
				}
			else
				{
					//assert(len < 4);
					if (len >= 4)
						*(int *)0 = 0;
					a = b = c = d = pad;
					for(i = 0; i < len; i++)
						{
							a <<= 8;
							a |= msg[i];
						}
				}
	
	TEACORE(FULLROUNDS);
	
/*	return 0;*/
	return h0^h1;
}


static u32 yura_hash (const char *msg, int len)
{
    int j, pow;
    u32 a, c;
    int i;
    
    for (pow=1,i=1; i < len; i++) pow = pow * 10; 
    
    if (len == 1) 
		a = msg[0]-48;
    else
		a = (msg[0] - 48) * pow;
    
    for (i=1; i < len; i++)
		{
			c = msg[i] - 48; 
			for (pow=1,j=i; j < len-1; j++) pow = pow * 10; 
			a = a + c * pow;
		}
    
    for (; i < 40; i++)
		{
			c = '0' - 48; 
			for (pow=1,j=i; j < len-1; j++) pow = pow * 10; 
			a = a + c * pow;
		}
    
    for (; i < 256; i++)
		{
			c = i; 
			for (pow=1,j=i; j < len-1; j++) pow = pow * 10; 
			a = a + c * pow;
		}
    
    a = a << 7;
    return a;
}

static u32 r5_hash (const char *msg, int len)
{
  u32 a=0;
  int i;

  for (i = 0; i < len; i ++)
	  {
		  a += msg[i] << 4;
		  a += msg[i] >> 4;
		  a *= 11;
	  }
  return a;
}


static struct  hash_function_t
{
	int  hash_function_code;
	hashf_t hash_function;
	char hash_function_name[10];
}  hash_function_table[]=
{
	TEA_HASH,   keyed_hash, "tea",
	YURA_HASH,  yura_hash,  "rupasov",
	R5_HASH,    r5_hash,     "r5",
};



static hashf_t hash_function (__u32 tt)
{
	int ret;
	ret=1;
	switch (tt)
		{
		case TEA_HASH:
		case YURA_HASH:
		case R5_HASH:
			ret=tt-1;
		};
//	printk ("Using %d %d %s hash to sort names\n" , tt,
//					hash_function_table[ret].hash_function_code, hash_function_table[ret].hash_function_name);
//			kbd_getc();
	return hash_function_table[ret].hash_function;
}


/* for debug */
static int vv_print_key(char *mes,struct key * k)
{
	printk("    %s  key= %d/%d/%d/%d\n",
		   mes,k->k_dir_id,k->k_objectid,k->k_offset,k->k_uniqueness);
			return kbd_getc();
}



/*make milo_item - copy unaligned space to aligned struct */
static struct milo_item * vv_m_ihead(struct milo_item * p_item, struct block_head * p_bh, int p_pos)
{

	memcpy( p_item, (char *)(p_bh)+BLKH_SIZE+IH_SIZE*(p_pos), IH_SIZE );

	memcpy( &(p_item->ih_r_key), &(p_bh->blk_right_delim_key), BLKH_SIZE);

	memcpy( &p_item->body, (char*)(p_bh)+ p_item->ih_item_location, p_item->ih_item_len );

	p_item->it_Number = p_pos;
	p_item->it_Last = p_pos == (p_bh->blk_nr_item-1) ;

	return p_item;
}


static int reiserfs_m_mount(long device, int quiet)
{
	extern void set_blocksize(kdev_t dev, int size);
	int i;
	long sb_offset;
     /* Initialize the inode table */
	for (i = 0; i < MAX_OPEN_FILES; i++)
		{
			inode_table[i].free = 1;
			inode_table[i].inumber = 0;
		}
	set_blocksize(device, REISERFS_BLOCK_SIZE);
	sb_offset = REISERFS_DISK_OFFSET_IN_BYTES;
	if (device_read(fd, (char *) &sb, sizeof(sb), sb_offset) !=  sizeof(sb))
		{
			printk("reiserfs sb read failed");
			device_close(fd);
			return (-1);
		}
	if (strncmp (sb.s_magic,  REISERFS_SUPER_MAGIC_STRING, strlen ( REISERFS_SUPER_MAGIC_STRING)))
		{
			printk ("reiserfs_read_super: can't find a reiserfs filesystem.\n");
			device_close(fd);		
			return -1;
		}
	blocksize = REISERFS_BLOCK_SIZE;
	reiserfs.blocksize = REISERFS_BLOCK_SIZE;
	p_hash_function = hash_function (sb.s_hash_function_code);
	return (0);
}


static struct milo_item * reiserfs_seach_item(struct key * p_key, char * p_buf, struct milo_item *p_item)
{
	int i, j;
	unsigned long offset;

	offset = blocksize * sb.s_root_block;
	bh = ( struct block_head *) p_buf;
	
    if ( device_read( fd, p_buf, blocksize, offset) !=blocksize )
		{
			printk("reiserfs_seach_item(read root of tree): read error\n");
			return (NULL);
		}
	
	while ( bh->blk_level > 1 )  /* Thearch leaf*/
		{
			i = 0;
			while( i <= VV_ITEMS(bh) && comp_keys( VV_KEY(bh,i), p_key )<0  )
				i++;
			
			if (comp_keys( VV_KEY(bh,i), p_key )>=0)
				i++;
			
			offset = VV_CHILD_NUM(bh,i);

			if ( device_read( fd, p_buf, blocksize, offset) !=blocksize )
				{
					printk("reiserfs_seach_item(next read tree): read error\n");
					return (NULL);
				}
		};

	j = 0;

	/* thearch item */
	while( j < VV_ITEMS(bh) && comp_keys((struct key *) vv_m_ihead( p_item, bh, j), p_key ) < 0 )
		j++;

	if ( comp_short_keys( (struct key *)(p_item), p_key ))
		{
			ErrorNumber = KEY_NOT_FOUND;
			if (j>0) vv_m_ihead( p_item, bh, j-1 );
		}
    return (p_item);	
}


static struct reiserfs_inode *reiserfs_m_iget( struct key * ino_k, struct milo_item * tmp_item)
{
    int i;
	struct reiserfs_inode *ip;
	struct inode_table_entry *itp;

	ip = NULL;
	itp = NULL;
    for (i = 0; i < MAX_OPEN_FILES; i++)
		{
			if (inode_table[i].free || inode_table[i].inumber==ino_k->k_objectid)
				{
					itp = &(inode_table[i]);
					ip = &(itp->inode);
					break;
				}
		}
	if ((ip == NULL) || (itp == NULL))
		{
			printk("reiserfs_m_iget: no free inodes\n");
			return (NULL);
		}
	
	itp->inumber = ino_k->k_objectid;
	itp->free = 0;
		
    ip->i_mode = 0;

    ip->i_key[0] = ino_k->k_dir_id;
	ip->i_key[1] = ino_k->k_objectid;
    ip->i_key[2] = SD_OFFSET;
    ip->i_key[3] = TYPE_STAT_DATA;

	reiserfs_seach_item( (struct key *)ip, &iblkbuf , tmp_item);
	
    ip->i_mode               = le16_to_cpu (tmp_item->body.sd.sd_mode);
    ip->i_uid                = le16_to_cpu (tmp_item->body.sd.sd_uid);
    ip->i_gid                = le16_to_cpu (tmp_item->body.sd.sd_gid);
    ip->i_size               = le32_to_cpu (tmp_item->body.sd.sd_size);
    ip->i_mtime              = le32_to_cpu (tmp_item->body.sd.sd_mtime);
    ip->i_atime              = le32_to_cpu (tmp_item->body.sd.sd_atime);
    ip->i_ctime              = le32_to_cpu (tmp_item->body.sd.sd_ctime);
	ip->i_blocks             = le32_to_cpu (tmp_item->body.sd.u.sd_blocks);
	ip->i_first_direct_byte  = le32_to_cpu (tmp_item->body.sd.sd_first_direct_byte);
	
	switch ( ip->i_mode & S_IFMT )
		{
		case S_IFREG:
			(ip->i_key[2])= 1;
			ip->i_key[3] = -1;
			reiserfs_seach_item( (struct key *)ip, &iblkbuf , tmp_item);
//			memcpy(&(ip->i_data),&(tmp_item->body),tmp_item->ih_item_len);
//			ip->i_data_length = tmp_item->ih_item_len/sizeof(__u32);
			break;
		case S_IFDIR:
			ip->i_key[2] = DOT_OFFSET;
			ip->i_key[3] = DIRENTRY_UNIQUENESS;
			reiserfs_seach_item( (struct key *)ip, &iblkbuf , tmp_item);
/*
//			memcpy(&(ip->i_data),&(tmp_item->body),tmp_item->ih_item_len);
//			ip->i_data_length = tmp_item->ih_item_len;
*/			
			break;
		case S_IFCHR:
			break;
		case S_IFBLK:
		case S_IFIFO:
		case S_IFSOCK:
			break;
		case S_IFLNK:
			reiserfs_seach_item( (struct key *)ip, &iblkbuf , tmp_item);
/*
//			memcpy(&(ip->i_data),&(tmp_item->body),tmp_item->ih_item_len);
//			ip->i_data_length = tmp_item->ih_item_len;
*/
			break;
		}
	return (ip);
}

/* Release our hold on an inode.  Since this is a read-only application,
 * don't worry about putting back any changes...
 */
static void reiserfs_m_iput(struct reiserfs_inode *ip)
{
	struct inode_table_entry *itp;

	/* Find and free the inode table slot we used... */
	itp = (struct inode_table_entry *) ip;

	itp->inumber = 0;
	itp->free = 1;
}

/* Return the size of the open file */
static long reiserfs_m_fdsize(int fd)
{
	struct reiserfs_inode *ip;

	ip = &(inode_table[fd].inode);
	return ip->i_size;
}

/* Read block number "blkno" from the specified file */
static int reiserfs_m_bread(int fd, long blkno, char *buffer)
{
	struct reiserfs_inode *ip;
	static struct inode_table_entry * itp;
	ip = &(inode_table[fd].inode);
	itp = ip;
	if (itp->free)
		{
			printk("file %d not open", fd);
			return -1;
		}
	reiserfs_m_breadi(ip, blkno, buffer);
	return 0;
}

#ifndef IOBUF_SIZE
#define IOBUF_SIZE 1024
#endif

static void reiserfs_m_breadi(struct reiserfs_inode *ip, long p_blkno, char *buffer)
{
	long dev_blkno,blkno;
	struct key pos_key;
	unsigned long offset;
	unsigned long offset_min;
	unsigned long offset_max;
	unsigned int	n_pos;
	unsigned long r_size;

	offset = p_blkno * IOBUF_SIZE   /*1024*/  ;

	if (offset > ip->i_size)
		{
			printk("\nattempt read %d belong end of file:%d",offset,ip->i_size);
			memset(buffer,0,blocksize);
			kbd_getc();
			return;
		}
	
	offset_min = it_open.ih_key.k_offset-1;
	offset_max = it_open.ih_item_len-1;

	if(it_open.ih_key.k_uniqueness== TYPE_INDIRECT )
			offset_max = offset_max * blocksize + offset_min;

	if ( offset < offset_min || offset > offset_max )
		{
			copy_key (&pos_key, ip);
			pos_key.k_offset = offset;
			reiserfs_seach_item( &pos_key, buffer, &it_open);
			vv_print_key(" bread: find new item",&pos_key);
		}

	switch(it_open.ih_key.k_uniqueness)
		{
		case TYPE_INDIRECT :
			offset -= (it_open.ih_key.k_offset - 1);
			n_pos  = offset / blocksize;
			offset = offset % blocksize;
			dev_blkno = it_open.body.item[n_pos];
			offset += dev_blkno * blocksize;
			if ( device_read (fd, buffer,   IOBUF_SIZE ,  offset) != IOBUF_SIZE )
				{
					printk("reiserfs_bread: read error");
				}
			break;
		case TYPE_DIRECT :
			memcpy( buffer, it_open.body.txt, it_open.ih_item_len );
			break;
		}
	return 0;
}


static __u32 m_get_third_component ( const char * name, int len)
{
	__u32 res;
	if (!len || (len == 1 && name[0] == '.'))           return DOT_OFFSET;
	if (len == 2 && name[0] == '.' && name[1] == '.')   return DOT_DOT_OFFSET;
	res = p_hash_function (name, len);
	res = GET_HASH_VALUE(res);
	if (res == 0) 	res = 128;
	return res + MAX_GEN_NUMBER;
}


static struct reiserfs_inode *reiserfs_m_namei(char *name, struct milo_item *n_item)
{
	char namebuf[512];
	char *component;
	struct reiserfs_inode
		*dir_inode = NULL,
		*file_inode;
	int i;
	int prev_offset;
struct reiserfs_de_head * de;
	

	struct key pos_key;	/* key of current position in the directory (key of directory entry) */
    struct key * rkey;
	int link_count;
	
	strcpy(namebuf, name);
	
  abs:
	file_inode =  reiserfs_m_iget ( &root_key, n_item);

  rel:
	if (!file_inode)
		return NULL;

	component = strtok(namebuf, "/");
	while (component)
		{
			int component_length;

			if (dir_inode)  reiserfs_m_iput(dir_inode);
			dir_inode = file_inode;
			file_inode = NULL;
			
			component_length = strlen(component);
			
//#ifdef MILO_SYMLINK_DIRS
			if (S_ISLNK(dir_inode->i_mode))
				{
					printk("     symlink     ");
					if (++link_count > 10)
						printk("%s: possibly circular symlink\n", name);
					else
						{
							char *t;
							reiserfs_seach_item( dir_inode, &iblkbuf, n_item);
							memcpy( blkbuf, n_item->body.txt, n_item->ih_item_len );
							t = blkbuf + strlen(blkbuf);
							while (component != NULL)
								{
									t = stpcpy(stpcpy(t, "/"), component);
									component = strtok(NULL, "/");
								}

							/* Put it back to namebuf */
							strcpy(namebuf, blkbuf);

							if (*namebuf == '/')
								{
									/* Absolute symlink */
									if (dir_inode)
										reiserfs_m_iput(dir_inode);
									if (file_inode)
										reiserfs_m_iput(file_inode);
									goto abs;
								}
							else
								{
									/* Relative symlink */
									if (file_inode)
										reiserfs_m_iput(file_inode);
									file_inode = dir_inode;
									goto rel;
								}
						}
				}
//#endif				/* MILO_SYMLINK_DIRS */
			
			if (!S_ISDIR(dir_inode->i_mode))
				{
					printk("%s: path component is not a directory\n", name);
					break;
				}

			copy_key (&pos_key, dir_inode);
			pos_key.k_offset = m_get_third_component( component, component_length);

			reiserfs_seach_item( (struct key *)dir_inode, &iblkbuf , n_item);

			de = n_item->body.deh;
			
			i = n_item->u.ih_entry_count;

			prev_offset = n_item->ih_item_len;
			
			while ( i  && GET_HASH_VALUE( de->deh_offset ) < GET_HASH_VALUE( pos_key.k_offset ) )
				{
					prev_offset = de->deh_location;
					de++;
					i--;
				}

			if ( strncmp( n_item->body.txt + de->deh_location , component, prev_offset - de->deh_location ) )
				{
					reiserfs_m_iput(dir_inode);
					return NULL;
				}
			pos_key.k_dir_id    = de->deh_dir_id;
			pos_key.k_objectid  = de->deh_objectid;
			pos_key.k_offset    = DOT_OFFSET;

			file_inode = reiserfs_m_iget(&pos_key,n_item);
			
			component = strtok(NULL, "/");
		}

	if (dir_inode)
		reiserfs_m_iput(dir_inode);
	return (file_inode);
}

/*
** read next byte
*/

static char vv_nex(long count, struct reiserfs_inode *ip,char * blkh, long f_size)
{
	unsigned int blkno;
	unsigned int offs;
	
	if ( count < f_size )
		{
			blkno = count / IOBUF_SIZE;
			offs  = count % IOBUF_SIZE;
			if ( !offs )
					reiserfs_m_breadi( ip, blkno, blkh);
			return (*(blkh+offs));
		}
	else return 0;
}


/*
** print file  
*/
static void reiserfs_m_cat(struct reiserfs_inode *ip, char * buf)
	/*char *filename, char * dev)*/
{
	long f_size;
	__u32 blkno;
	int fd;
	int i,j;
	char * blkh;
	char * component;
	struct inode_table_entry * itp;
	char cc;
	reiserfs_seach_item( ip, buf, &it_open);

	blkh = buf;

	itp = (struct inode_table_entry *) ip;
	fd = (itp - inode_table);

	f_size=reiserfs_m_fdsize(fd);
	printk( "file size=%d\n", f_size);
	blkno=0;
	i=0;
	j=0;
	printk("\n%3d:",++i);
	while( cc = vv_nex( j++, ip, blkh, f_size))
		{
			printk( "%c", cc );
			if ( cc == '\n' )
				{
					if( !( i%20 ) )
						if ( kbd_getc() == 'q' )
							break;
					printk( "%3d:", ++i );
				}
		}

	printk("\n");
	reiserfs_m_close(fd);
}

/*
** ls (*dirname, void * devname)
** devname - not use, no need: milo read from 1 device, opened befor
** find dirname, and list it contains.
**   if finded name is file- trie print it
*/
static void reiserfs_m_ls(char *dirname, char *devname)
{
	char filename[512];
	int filenamelen;
	char format_string[16];
	struct reiserfs_inode *dir_inode;
	struct reiserfs_inode *file_inode;
	int columns;
	int col;
	char typetag;
	struct block_head * blkh;
	struct reiserfs_de_head * de;

	struct key pos_key;	/* key of current position in the directory (key of directory entry) */
    struct key * rkey;
    char small_buf[REISERFS_BLOCK_SIZE] ;
	int i;
	__u32 prev_loc;

#define NOT_EOD NotEOD
	int NotEOD;

	strcpy(format_string,"   %s");

	/* Get the inode for the specified directory */
	dir_inode = reiserfs_m_namei(dirname,&it_buf);

	printk("\nls %s",dirname);
	
	if (!dir_inode)
		{
			printk("%s: is no  file or directory\n", dirname);
			return;
		}

	/* If this inode exists but is *not* a directory, just cat it here */
	if ((dir_inode->i_mode & S_IFMT) != S_IFDIR)
		{
			reiserfs_m_cat(dir_inode, &it_buf);
			return;
		}

	col = 0;
	columns = 4;
	NOT_EOD = 1;
	blkh = &small_buf;
	while ( NOT_EOD )
		{
			NOT_EOD = 0;

			reiserfs_seach_item( dir_inode, blkh, &it_buf);
			de = it_buf.body.deh;
			i = it_buf.u.ih_entry_count;
			prev_loc = it_buf.ih_item_len;
			printk("\ndirectory %d entries:",i);

			while ( i  )  /* loop by dir_item*/
				{
					filenamelen          = prev_loc - de->deh_location;
					memcpy( filename, it_buf.body.txt + de->deh_location ,filenamelen);
					filename[filenamelen] = '\0';
					pos_key.k_dir_id     = de->deh_dir_id;
					pos_key.k_objectid   = de->deh_objectid;
					pos_key.k_offset     = m_get_third_component( filename, filenamelen );
					pos_key.k_uniqueness = DIRENTRY_UNIQUENESS;
					file_inode = reiserfs_m_iget(&pos_key,&n_item_buf);
					if (file_inode)
						{
							switch (file_inode->i_mode & S_IFMT)
								{
								case S_IFLNK:
									typetag = '@';
									break;
								case S_IFDIR:
									typetag = '/';
									break;
								case S_IFIFO:
								case S_IFSOCK:
									typetag = '=';
									break;
								default:
									typetag = ' ';
									break;
								}
							reiserfs_m_iput(file_inode);
						}
					else
						{
							typetag = '!';
						}
					filename[filenamelen] = typetag;
					filename[filenamelen + 1] = '\0';

					printk(format_string, filename);
					
					if ((++col % columns) == 0)  printk("\n");
					
					prev_loc = de->deh_location;

					de++;
					i--;
				}
			if ( it_buf.it_Last ) /* This item is the last in block */
				{
					if ( !COMP_SHORT_KEYS (&(it_buf.ih_r_key), &pos_key))
						{
							/* directory continues in the right neighboring block */
							copy_key (&pos_key, &(it_buf.ih_r_key));
							NOT_EOD = 1;
						}
					
				}
		}			/* end-while */
	
	if ((col % columns) != 0)
		{
			printk("\n");
		}

	/* We're done with the directory now... */
	reiserfs_m_iput(dir_inode);

}

static int reiserfs_m_open(char *filename)
{
	/* Unix-like open routine.  Returns a small integer (actually an index into
	 * the inode table... */
	struct reiserfs_inode *ip;
	ip = reiserfs_m_namei(filename,&it_open);

	if (ip)
		{
			struct inode_table_entry *itp;

			reiserfs_seach_item( ip, &n_item_buf, &it_open);
/*
			printk("\n open: %s size:%d/%d itemsize:%d fist block:%d",
				   filename, ip->i_size, ip->i_blocks, it_open.ih_item_len,it_open.body.item[0]);
*/			
			itp = (struct inode_table_entry *) ip;
			return (itp - inode_table);
		}
	else
		{
			printk("not opened");
			return (-1);
		}
}


static void reiserfs_m_close(int fd)
{
	reiserfs_m_iput(&(inode_table[fd].inode));
	device_close(fd);
}
